import logging

from contiamo.http_client import HTTPClient
from contiamo.utils import raise_response_error, parse_query_result


logger = logging.getLogger(__name__)


def CreateNestedResource(base_class, parent, **kwargs):
    class_name = type(parent).__name__ + base_class.__name__.rstrip('Resource')
    properties = {'parent': parent}
    properties.update(kwargs)
    return type(class_name, (base_class,), properties)


###
# Main client
###
class Client:

    def __init__(self, api_key, api_base='https://api.contiamo.com'):
        self.api_key = api_key
        self.api_base = api_base
        self.http_client = HTTPClient(api_key)
        self.Project = CreateNestedResource(ProjectResource, parent=self, path_segment=None)

    def client(self):
        return self.http_client

    def instance_url(self):
        return self.api_base


###
# Base resource class
###
class Resource(dict):
    id_attribute = 'id'
    path_segment = None

    def __init__(self, id):
        self.id = id
        self._init_nested_resources()

    def _init_nested_resources(self):
        pass

    @classmethod
    def client(cls):
        return cls.parent.client()

    @classmethod
    def class_url(cls):
        base_url = cls.parent.instance_url()
        if cls.path_segment:
            return '%s/%s' % (base_url, cls.path_segment)
        else:
            return base_url

    def instance_url(self):
        return '%s/%s' % (self.class_url(), self.id)

    @classmethod
    def instantiate_from_response(cls, response):
        try:
            resource_instance = cls(response[cls.id_attribute])
        except KeyError as e:
            raise_response_error(e, response, logger)
        resource_instance.update(response)
        return resource_instance

    @classmethod
    def request(cls, method, url, payload=None):
        response = cls.client().request(method, url, payload=payload)
        try:
            result = response.json()
        except ValueError as e:  # JSONDecodeError inherits from ValueError
            raise_response_error(e, response, logger)
        return result

    def resolve_url(self, url, sub_path):
        resolved_url = url if url else self.instance_url()
        if sub_path is not None:
            resolved_url += '/' + sub_path.strip('/')
        return resolved_url

    # do not overwrite dict get method
    def _get(self, url=None, sub_path=None):
        return self.__class__.request('GET', self.resolve_url(url, sub_path))

    def _post(self, url=None, sub_path=None, payload=None):
        return self.__class__.request('POST', self.resolve_url(url, sub_path), payload=payload)

    def _put(self, url=None, sub_path=None, payload=None):
        return self.__class__.request('PUT', self.resolve_url(url, sub_path), payload=payload)


###
# Mixin classes
###
class RetrievableResource(Resource):

    @classmethod
    def instantiate_list(cls, responses):
        return [cls.instantiate_from_response(response) for response in responses]

    @classmethod
    def list(cls, instantiate=False):
        response = cls.request('get', cls.class_url())
        # resources can be returned as list or dictionary
        if type(response) is list:
            resources = response
        else:
            try:
                resources = response['resources']
            except (TypeError, KeyError) as e:
                raise_response_error(e, response, logger)
        if instantiate:
            return cls.instantiate_list(resources)
        else:
            return resources

    @classmethod
    def retrieve(cls, id):
        instance = cls(id)
        instance.fetch()
        return instance

    def fetch(self):
        self.update(self._get())


class QueryableResource(Resource):

    def data(self):
        return self._post(sub_path='data')


class UpdateableResource(Resource):

    @classmethod
    def create(cls, model):
        response = cls.request('post', cls.class_url(), payload=model)
        return cls.instantiate_from_response(response)

    def modify(self, model):
        # need to handle lock version
        response = self._put(payload=model)
        return self.instantiate_from_response(response)

    def delete(self):
        response = self.request('DELETE', self.instance_url())
        return response


###
# Resource classes
###
class ProjectResource(Resource):
    path_segment = 'projects'

    def _init_nested_resources(self):
        self.Dashboard = CreateNestedResource(DashboardResource, parent=self)
        self.App = CreateNestedResource(AppResource, parent=self)
        self.PublishedQuery = CreateNestedResource(PublishedQuery, parent=self)

    def query_sql(self, app_id, sql):
        payload = {
            'app_data_id': app_id,
            'columns': [],
            'query': sql
        }
        json_response = self._post(sub_path='/sql_query', payload=payload)
        # SQL queries do not return 'date' as a data type, so there is no point in parsing dates in the query result.
        # However, we could parse the date in field_date_raw afterwards.
        return parse_query_result(json_response, parse_dates=False, use_column_names=False)

    def query(self, payload, parse_dates=True, use_column_names=True):
        json_response = self._post(sub_path='/query', payload=payload)
        return parse_query_result(json_response, parse_dates=parse_dates, use_column_names=use_column_names)


class DashboardResource(RetrievableResource, UpdateableResource):
    path_segment = 'dashboards'

    def _init_nested_resources(self):
        self.Widget = CreateNestedResource(WidgetResource, parent=self)


class WidgetResource(RetrievableResource, QueryableResource, UpdateableResource):
    path_segment = 'widgets'


class AppResource(RetrievableResource):
    path_segment = 'apps'

    def _init_nested_resources(self):
        self.Contract = CreateNestedResource(ContractResource, parent=self)


class ContractResource(RetrievableResource, UpdateableResource):
    path_segment = 'data_contracts/contracts'
    id_attribute = 'key'


class PublishedQuery(QueryableResource):
    path_segment = 'published_queries'
    name = 'Query'



import itertools
import os

import numpy as np
import pandas as pd
import pytest
import vcr
import yaml

from . import utils
from contiamo.data import DataClient, select_date_columns, select_int_columns, preformat, slice_in_chunks
from contiamo.errors import InvalidRequestError
from contiamo.public import query


try:
    config = yaml.safe_load(open(utils.file_test_data('test_config.yml')))
except OSError:
    raise RuntimeError("Test config file not found. Email brandon@contiamo.com.")

api_base = config['api_base']
contract_id = config['contract_id']
contract_token = config['contract_token']
query_id = config['query_id']

data_client = DataClient(contract_id, contract_token, api_base=api_base)


class TestRequests:

    def test_select_dates(self):
        df = pd.DataFrame(
            {'date': pd.date_range('2016-01-01', periods=2)})
        df['datetime'] = pd.to_datetime(df['date'].astype(str) + 'T01:00')
        df.loc[3] = None  # test NaT datetime
        assert select_date_columns(df) == ['date']
        df = pd.DataFrame(
            {'int': [0, 1], 'float': [0.1, 0.2], 'str': ['a', 'b']})
        assert select_date_columns(df) == []

    def test_select_integers(self):
        df = pd.DataFrame({
            'int': list(range(4)),
            'int2': [3, np.nan, 58, np.nan],
            'int3': [2, 57, np.nan, np.nan],
            'float': [2.5, 2.6, 2.7, 2.0],
        })
        assert set(select_int_columns(df)) == {'int2', 'int3'}

    def test_preformat(self):
        df = pd.DataFrame({
            'int': [3, np.nan, 58],
            'date': pd.date_range('2016-01-01', periods=3),
        })
        preformat(df)
        expected = pd.DataFrame({
            'int': ['3', np.nan, '58'],
            'date': ['2016-01-01', '2016-01-02', '2016-01-03'],
        })
        assert df.equals(expected)

    @vcr.use_cassette(utils.file_test_cassette('test_purge.yaml'))
    def test_purge(self):
        response = data_client.purge()
        assert 'status' in response
        assert response['status'] == 'ok'

    @vcr.use_cassette(utils.file_test_cassette('test_discover.yaml'))
    def test_discover(self):
        # need to purge data before discovering
        if not os.path.isfile(utils.file_test_cassette('test_discover.yaml')):
            data_client.purge()
        response = data_client.discover(filename=utils.file_test_data('mock_data.csv'))
        assert 'status' in response
        assert response['status'] == 'ok'

    @vcr.use_cassette(utils.file_test_cassette('test_upload.yaml'))
    def test_upload(self):
        response = data_client.upload(filename=utils.file_test_data('mock_data.csv'))
        assert 'status' in response
        assert response['status'] == 'ok'

    @vcr.use_cassette(utils.file_test_cassette('test_upload.yaml'))
    def test_upload_dataframe(self):
        df = pd.DataFrame({'a': [1, 3], 'b': [2, 4]})
        response = data_client.upload(dataframe=df)
        assert 'status' in response
        assert response['status'] == 'ok'

    @vcr.use_cassette(utils.file_test_cassette('test_query.yaml'))
    def test_query(self):
        response = query(query_id, api_base=api_base)
        assert isinstance(response, pd.DataFrame)

    def test_slice_in_chunks(self):
        def reconstitute_list(l, slices):
            return list(itertools.chain.from_iterable((l[sl] for sl in slices)))
        for length, chunk_size in zip([113, 100, 100, 1, 0], [7, 100, 1, 10, 10]):
            data = list(range(length))  # we test slicing on a list
            assert data == reconstitute_list(data, slice_in_chunks(len(data), chunk_size))


class TestErrors:

    def file_test_dataframe(self):
        df = {'a': [1, 2, 3], 'b': [4, 5, 6]}
        with pytest.raises(InvalidRequestError):
            data_client.upload(dataframe=df)

    def test_no_argument(self):
        with pytest.raises(InvalidRequestError):
            data_client.upload()

    def test_two_arguments(self):
        df = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6]})
        with pytest.raises(InvalidRequestError):
            data_client.upload(dataframe=df, filename=utils.file_test_data('mock_data.csv'))



import os


def file_test_data(filename):
    return os.path.join(os.path.dirname(__file__), 'data', filename)


def file_test_cassette(filename):
    return os.path.join(os.path.dirname(__file__), 'cassettes', filename)



import logging

import requests
from requests.exceptions import RequestException

from contiamo import errors


logger = logging.getLogger(__name__)


class HTTPClient:

    def __init__(self, api_key=None):
        self.api_key = api_key

    def request(self, method, url, headers=None, payload=None, **kwargs):
        logger.debug('Sending %s request to %s' % (method.upper(), url))
        if payload and logger.isEnabledFor(logging.DEBUG):
            logger.debug('With payload: ' + str(payload))

        if headers is None:
            headers = {}
        headers.update({'Accept': 'application/json'})
        if self.api_key:
            headers.update({'X-API-TOKEN': self.api_key})

        try:
            response = requests.request(
                method,
                url,
                headers=headers,
                json=payload,
                **kwargs)
        except (TypeError, RequestException) as e:
            self._handle_request_error(e)

        if response.status_code < 200 or response.status_code >= 300:
            self._handle_api_error(response)

        return response

    def _handle_request_error(self, e):
        if isinstance(e, TypeError):
            raise TypeError(
                'You most likely have an outdated version of the requests library. Try upgrading to the latest '
                'version.\nThe error message was: %s' % e)
        else:
            # we may need to handle all exceptions, not just requests ones
            logger.error('%s connection error: %s' % (type(e).__name__, e))
            raise errors.APIConnectionError(
                'The connection attempt raised the following %s error:\n%s' % (type(e).__name__, e))

    def _handle_api_error(self, response):
        logger.error('HTTP error %d: %s' %
                     (response.status_code, response.text))

        # check for authorization error first
        is_auth_error = False
        try:
            is_auth_error = not response.json()['logged_in']
        except (ValueError, KeyError):
            pass
        if is_auth_error:
            raise errors.AuthenticationError(
                http_body=response.content, http_status=response.status_code, headers=response.headers)

        ERROR_MAPPING = {
            404: errors.NotFoundError,
            412: errors.DataSourceError,
            410: errors.QueryError,
            424: errors.QueryError,
            409: errors.UpdateError,
            422: errors.UpdateError,
        }
        try:
            error_class = ERROR_MAPPING[response.status_code]
        except KeyError:
            error_class = errors.APIError  # catch all
        raise error_class(http_body=response.content, http_status=response.status_code, headers=response.headers)



import datetime
import logging
import tempfile
import warnings

import numpy
import pandas as pd

from contiamo.errors import ContiamoException, InvalidRequestError
from contiamo.http_client import HTTPClient
from contiamo.utils import contract_url_template_from_identifier, raise_response_error, get_file_extension


logger = logging.getLogger(__name__)


ALLOWED_UPLOAD_FILETYPES = {'csv', 'jsonl'}


def select_date_columns(df):
    """Indentify datetime columns with dates only."""
    date_cols = []
    for col in df.select_dtypes(include=['datetime']).columns:
        if (df[col].dropna().dt.time == datetime.time(0)).all():
            date_cols.append(col)
    return date_cols


def select_int_columns(df):
    """Identify float columns with integer values only."""
    int_cols = []
    for col in df.select_dtypes(include=['float']).columns:
        if numpy.isclose(df[col].round(0), df[col], equal_nan=True).all():
            int_cols.append(col)
    return int_cols


def preformat(df):
    """
    Preprocess to improve default JSON serialization.

    - dates: serialize without time when represented as datetime
    - integers: serialize as string when represented as float
    """
    # dates
    date_cols = select_date_columns(df)
    if date_cols:
        df[date_cols] = df[date_cols].apply(
            lambda column: column.dt.strftime('%Y-%m-%d'))
    # integers
    int_cols = select_int_columns(df)
    for col in int_cols:
        # pandas assignment by index will fill NaN values back in their original locations
        df[col] = df.loc[df[col].notnull(), col].astype(int).astype(str)


def slice_in_chunks(length, chunk_size):
    if chunk_size < 1:
        raise InvalidRequestError('Chunk size cannot be less than 1.')
    nb_chunks = (length - 1) // chunk_size + 1
    return [slice(n * chunk_size, (n+1) * chunk_size) for n in range(nb_chunks)]


class DataClient:

    def __init__(self, contract_id, token, api_base='https://api.contiamo.com'):
        self.api_base = api_base
        self.token = token
        self.url_template = contract_url_template_from_identifier(
            contract_id, api_base)
        self.client = HTTPClient()

    def _make_request(self, url, file_object=None, file_type=None):
        kwargs = {'params': {'resource_token': self.token}}

        if file_object:
            if not file_type:
                raise ValueError(
                    'file_type must be specified if file_object is present')

            param_name = 'file-json' if file_type == 'jsonl' else 'file'
            kwargs.update({'files': {param_name: file_object}})

        response = self.client.request('post', url, **kwargs)
        try:
            json_response = response.json()
        except ValueError as e:  # JSONDecodeError inherits from ValueError
            raise_response_error(e, response, logger)

        return json_response

    def post_df(self, url, dataframe, include_index=False):
        if include_index:
            dataframe = dataframe.reset_index(drop=False)

        with tempfile.NamedTemporaryFile() as f:
            dataframe.to_json(f.name, orient='records',
                              lines=True, date_format='iso', date_unit='s')
            f.seek(0)
            result = self._make_request(url, f, file_type='jsonl')
        return result

    def post_file(self, url, filename):
        extension = get_file_extension(filename).lower()

        if extension not in ALLOWED_UPLOAD_FILETYPES:
            raise InvalidRequestError(
                'Unsupported file type \'%s\' to upload. Allowed formats: %s' %
                (extension, ', '.join(ALLOWED_UPLOAD_FILETYPES))
            )

        with open(filename, 'rb') as f:
            result = self._make_request(url, f, file_type=extension)
        return result

    def _post_data(self, url, dataframe, filename, include_index):
        # sanity checks
        if dataframe is None and filename is None:
            raise InvalidRequestError(
                'You need to specify at least a dataframe or a file to upload.')
        if dataframe is not None and filename is not None:
            raise InvalidRequestError(
                'Ambiguous request: You cannot provide both a dataframe and a file to upload.')
        if dataframe is not None and not isinstance(dataframe, pd.DataFrame):
            raise InvalidRequestError(
                'The argument you passed is a %s, not a pandas dataframe.'
                % type(dataframe).__name__)

        if dataframe is not None:
            return self.post_df(url, dataframe, include_index)
        else:
            return self.post_file(url, filename)

    ###
    # Public methods

    # Any dataframe passed to these methods is copied so it will not be modified.
    # Private methods can therefore modify the dataframe they receive.
    # We can address memory issues by adding a flag later on.

    def discover(self, dataframe=None, filename=None, include_index=False):
        if dataframe is not None:
            # no need to upload full dataframe for discovery
            dataframe = dataframe.sample(min(100, len(dataframe)))
            preformat(dataframe)
        url = self.url_template.format(action='upload/discover')
        return self._post_data(url, dataframe, filename, include_index)

    def upload(self, dataframe=None, filename=None, include_index=False, chunk_size=None):
        """Upload data from dataframe or file to data contract.

        - dataframe: pandas dataframe to be uploaded
        - filename: name of CSV or JSONL file containing the data to upload
            (file extension must be .csv or .jsonl, case insensitive)
        - include_index: if True, the index of the dataframe will be uploaded as well
        - chunk_size: if integer >= 1 is specified, the data will be uploaded in chunks
            (100k is a good default value for large data sets)
        """
        if dataframe is not None:
            dataframe = dataframe.copy()
        url = self.url_template.format(action='upload/process')

        # if dataframe is above chunk_size, upload in chunks
        if dataframe is not None and chunk_size is not None and len(dataframe) > chunk_size:
            chunk_slices = slice_in_chunks(len(dataframe), chunk_size)
            for idx, sl in enumerate(chunk_slices):
                try:
                    # we ignore the response, all that matters for uploads is the response code
                    # anything other than a 200 will raise an error
                    self._post_data(url, dataframe.iloc[sl], filename, include_index)
                except ContiamoException as e:
                    warnings.warn(
                        'Request #%d of %d has failed, %d rows have been uploaded so far.'
                        % (idx+1, len(chunk_slices), idx*chunk_size))
                    raise e
            # if no errors have been raised, simulate a response:
            return {'status': 'ok', 'requests_sent': len(chunk_slices)}

        return self._post_data(url, dataframe, filename, include_index)

    def purge(self):
        url = self.url_template.format(action='data_purge')
        return self._make_request(url)



import responses
import unittest

from . import utils
from contiamo.resources import Client
from contiamo.public import query
from contiamo.data import DataClient
from contiamo.utils import query_url_from_identifier, contract_url_template_from_identifier
from contiamo.errors import (
    APIConnectionError, APIError, AuthenticationError, DataSourceError, NotFoundError, QueryError, ResponseError,
    UpdateError)


# the only meaningful parameter is api_base, which should not resolve
api_base = 'http://xyz.wrong-domain-name.123'
resource_id = '456'
contiamo_client = Client('some_api_key', api_base=api_base)
project = contiamo_client.Project(id='123')
resource = project.Dashboard(resource_id)

# resources module


def mock_erroneous_request(method, response_body, status, payload=None):
    url = resource.class_url() if method == 'create' else resource.instance_url()
    verb = {'retrieve': responses.GET, 'create': responses.POST, 'modify': responses.PUT}.get(method)
    responses.add(verb, url, body=response_body, status=status, content_type='application/json')

    argument = payload if payload else resource_id
    request_method = getattr(resource, method)
    request_method(argument)


def mock_erroneous_retrieve(response_body, status):
    mock_erroneous_request('retrieve', response_body, status)


def mock_erroneous_create(response_body, status):
    mock_erroneous_request('create', response_body, status, payload={'name': 'ResourceName'})


def mock_erroneous_update(response_body, status):
    mock_erroneous_request('modify', response_body, status, payload={'name': 'AnotherName'})

# public module


def mock_erroneous_query(response_body, status):
    query_id = 'query:olap:48590200:21237:randomquerytoken'
    query_url, token = query_url_from_identifier(query_id, api_base)
    responses.add(responses.GET, query_url, body=response_body, status=status, content_type='application/json')
    query(query_id, api_base=api_base)

# data module


def mock_erroneous_upload(response_body, status):
    contract_id = 'contract:48590121:666570779:test'
    contract_url = contract_url_template_from_identifier(contract_id, api_base).format(action='upload/process')
    responses.add(responses.POST, contract_url, body=response_body, status=status, content_type='application/json')
    data_client = DataClient(contract_id, 'contract_token', api_base=api_base)
    data_client.upload(filename=utils.file_test_data('mock_data.csv'))


class ErrorTestCase(unittest.TestCase):

    def test_connection_error(self):
        with self.assertRaises(APIConnectionError):
            resource.retrieve(resource_id)

    @responses.activate
    def test_auth_error(self):
        with self.assertRaises(AuthenticationError):
            mock_erroneous_retrieve('{"error":"not_found","logged_in":false}', 404)
        with self.assertRaises(AuthenticationError):
            mock_erroneous_retrieve('{"error":"server_error","logged_in":false}', 500)

    @responses.activate
    def test_notfound_error(self):
        with self.assertRaises(NotFoundError):
            mock_erroneous_retrieve('{"error":"not_found","logged_in":true}', 404)

    @responses.activate
    def test_api_error(self):
        with self.assertRaises(APIError):
            mock_erroneous_retrieve('{}', 500)

    @responses.activate
    def test_create_error(self):
        with self.assertRaises(UpdateError):
            mock_erroneous_create('{}', 422)

    @responses.activate
    def test_update_error(self):
        with self.assertRaises(UpdateError):
            mock_erroneous_update('{}', 409)
        with self.assertRaises(UpdateError):
            mock_erroneous_update('{}', 422)

    @responses.activate
    def test_invalid_response(self):
        with self.assertRaises(ResponseError):
            mock_erroneous_retrieve('{"invalid":"json"', 200)
        with self.assertRaises(ResponseError):
            mock_erroneous_upload('{"invalid":"json"', 200)
        with self.assertRaises(ResponseError):
            mock_erroneous_query('{"invalid":"json"', 200)

    @responses.activate
    def test_data_source_error(self):
        with self.assertRaises(DataSourceError):
            mock_erroneous_upload('{}', 412)
        with self.assertRaises(DataSourceError):
            mock_erroneous_query('{}', 412)

    @responses.activate
    def test_query_error(self):
        with self.assertRaises(QueryError):
            mock_erroneous_query('{}', 410)
        with self.assertRaises(QueryError):
            mock_erroneous_query('{}', 424)

    @responses.activate
    def test_include_error(self):
        error_message = 'This is an error message.'
        with self.assertRaises(UpdateError) as cm:
            mock_erroneous_update('{"error": "%s"}' % error_message, 409)
        self.assertTrue(error_message in str(cm.exception))


if __name__ == '__main__':
    unittest.main()



import unittest
import json
import datetime

from . import utils
from contiamo.utils import parse_query_result as parse_json


class ParseJsonTestCase(unittest.TestCase):

    def test_simple(self):
        with open(utils.file_test_data('data_simple.json')) as data_file:
            result = json.load(data_file)
        df = parse_json(result)
        self.assertAlmostEqual(df['New Users'].iloc[0], 7447.0, 6)

    def test_hourly(self):
        with open(utils.file_test_data('data_hourly.json')) as data_file:
            result = json.load(data_file)
        df = parse_json(result)
        self.assertEqual(df['Date'].iloc[-1], datetime.datetime(2015, 12, 9, 23))

    def test_weekly(self):
        with open(utils.file_test_data('data_weekly.json')) as data_file:
            result = json.load(data_file)
        df = parse_json(result)
        self.assertEqual(df['Date'].iloc[0], datetime.datetime(2015, 11, 23))

    def test_no_parse(self):
        with open(utils.file_test_data('data_weekly.json')) as data_file:
            result = json.load(data_file)
        df = parse_json(result, parse_dates=False)
        self.assertEqual(df['Date'].iloc[0], '2015W48')

    def test_dimension(self):
        with open(utils.file_test_data('data_dimension.json')) as data_file:
            result = json.load(data_file)
        df = parse_json(result)
        self.assertEqual(df['Default Channel Grouping'].iloc[0], 'Direct')

    def test_date_dimension(self):
        with open(utils.file_test_data('data_date_dimension.json')) as data_file:
            result = json.load(data_file)
        df = parse_json(result)
        self.assertEqual(df['First purchase quarter'].iloc[0], datetime.datetime(2013, 7, 1))

    def test_nodate(self):
        with open(utils.file_test_data('data_nodate.json')) as data_file:
            result = json.load(data_file)
        df = parse_json(result)
        self.assertAlmostEqual(df['No. of orders'].iloc[0], 237.0, 6)

    def test_stat(self):
        with open(utils.file_test_data('data_stat.json')) as data_file:
            result = json.load(data_file)
        df = parse_json(result)
        self.assertAlmostEqual(df['Pageviews (Exponential Moving Average)'].iloc[0], 17562.0, 6)

    def test_calc(self):
        with open(utils.file_test_data('data_calc.json')) as data_file:
            result = json.load(data_file)
        df = parse_json(result)
        self.assertAlmostEqual(df['Calculation'].iloc[0], 1156540.0, 6)

    def test_comparison_period(self):
        with open(utils.file_test_data('data_comparison_period.json')) as data_file:
            result = json.load(data_file)
        df = parse_json(result)
        self.assertAlmostEqual(df['Pageviews (previous period)'].iloc[0], 19302.0, 6)

    def test_empty(self):
        with open(utils.file_test_data('data_empty.json')) as data_file:
            result = json.load(data_file)
        df = parse_json(result)
        self.assertEqual(len(df), 0)

    def test_duplicate_columns(self):
        with open(utils.file_test_data('data_duplicate_columns.json')) as data_file:
            result = json.load(data_file)
        df = parse_json(result)
        self.assertEqual(len(df.columns), 4)
        self.assertTrue(df.columns.is_unique)

    def test_duplicate_calculation(self):
        with open(utils.file_test_data('data_duplicate_calculation.json')) as data_file:
            result = json.load(data_file)
        df = parse_json(result)
        self.assertEqual(len(df.columns), 4)
        self.assertTrue(not df.columns.is_unique)


if __name__ == '__main__':
    unittest.main()



import os
import pandas as pd

from contiamo.dateparser import DateParser


###
# parse labs ids
def _raise_invalid_identifier_error(identifier, e):
    from contiamo.errors import InvalidRequestError
    raise InvalidRequestError(
        'The identifier "%s" is invalid. The following error was raised.\n'
        '%s: %s' % (identifier, type(e).__name__, e))


query_url_template = '{api_base}/{project_id}/published_queries/{query_id}/data.json'


def query_url_from_identifier(query_identifier, api_base):
    """query:query_type:project_id:query_id:token"""
    try:
        _, project_id, query_id, token = query_identifier.split(':')[1:]
    except ValueError:
        # backward compatibility
        try:
            project_id, query_id, token = query_identifier.split(':')[1:]
        except ValueError as e:
            _raise_invalid_identifier_error(query_identifier, e)

    url = query_url_template.format(
        api_base=api_base,
        project_id=project_id,
        query_id=query_id)

    return url, token


contract_url_template = '{api_base}/{project_id}/apps/{app_id}/data_contracts/contracts/{contract_key}/{{action}}'
# escape curly braces to keep {action} in the template


def contract_url_template_from_identifier(contract_identifier, api_base):
    """contract:project_id:app_id:contract_key"""
    try:
        project_id, app_id, contract_key = contract_identifier.split(':')[1:]
    except ValueError as e:
        _raise_invalid_identifier_error(contract_identifier, e)

    return contract_url_template.format(
        api_base=api_base,
        project_id=project_id,
        app_id=app_id,
        contract_key=contract_key,
    )


###
# parse query response
def raise_json_error(e, response):
    from contiamo.errors import ResponseError
    raise ResponseError(
        'The JSON response from the server was invalid. Please report the bug to support@contiamo.com\n'
        'The following %s error was raised when parsing the JSON response:\n%s' % (
            type(e).__name__, e),
        json_body=response)


def parse_query_result(json_response, parse_dates=True, use_column_names=True):
    # The function below is not safe (only works with floats) so we keep it in the local namespace.
    def is_integer(a):
        return a.is_integer()

    df = pd.DataFrame()
    try:
        for idx, column in enumerate(json_response['columns']):
            col_key = column['key']
            df[col_key] = [row[idx] for row in json_response['rows']]
            # Format
            if column['data_type'] == 'numeric':
                df[col_key] = pd.to_numeric(df[col_key])
                # try converting to integer if applicable
                try:
                    if df[col_key].dtype == 'float':
                        if df[col_key].map(is_integer).all():
                            df[col_key] = df[col_key].astype(int)
                except Exception:
                    pass  # really, we do not want this workaround to create any errors if it fails
            if parse_dates and column['data_type'] == 'date' and len(df[col_key]) > 0:
                parser = DateParser()
                parser.identifyPeriodUnit(df[col_key][0])
                df[col_key] = df[col_key].map(parser.parse)

        # once the dataframe is built with keys, overwrite column names if requested
        if use_column_names:
            column_names = []
            for column in json_response['columns']:
                if 'name' in column:
                    column_names.append(column['name'])
                else:
                    # column is a date, but can be the comparison date
                    if column.get('comparison'):
                        column_names.append('Date (comparison)')
                    else:
                        column_names.append('Date')
            if len(column_names) > len(set(column_names)):
                # column names not unique: use app_data name for metrics
                for idx, column in enumerate(json_response['columns']):
                    if column['column_type'] == 'metric':
                        try:
                            column_names[idx] = column['app_data']['name'] + \
                                ': ' + column['name']
                        except KeyError:
                            # calculations do not have an app_data
                            pass
            df.columns = column_names

    except KeyError as e:
        raise_json_error(e, json_response)
    return df

###
# misc


def raise_response_error(e, response, logger=None):
    from contiamo.errors import ResponseError
    if logger:
        logger.error('Invalid JSON response: %s' % response.text)
    raise ResponseError(
        'The JSON response from the server was invalid. Please report the bug to support@contiamo.com\n'
        'The following %s error was raised when parsing the JSON response:\n%s' % (
            type(e).__name__, e),
        http_body=response.content, http_status=response.status_code, headers=response.headers)


def get_file_extension(filename):
    _, ext = os.path.splitext(filename)
    return ext.lstrip('.')



import unittest

from contiamo.resources import CreateNestedResource, Client, Resource


class ClassNestingTestCase(unittest.TestCase):
    api_base = 'https://api.some-name.com'
    project_id = '123456'
    dashboard_id = '456789'

    def setUp(self):
        self.contiamo_client = Client('token', api_base=self.api_base)
        self.project = self.contiamo_client.Project(self.project_id)
        self.dashboard = self.project.Dashboard(self.dashboard_id)

    def test_nested_urls(self):
        self.assertEqual(self.project.instance_url(), '%s/%s' %
                         (self.api_base, self.project_id))
        self.assertEqual(self.dashboard.instance_url(), '%s/%s/%s' %
                         (self.project.instance_url(), 'dashboards', self.dashboard_id))

    def test_client(self):
        self.assertIs(self.dashboard.client(), self.contiamo_client.client())

    def test_nested_class_name(self):
        class ChildResource(Resource):
            pass

        class Parent(Resource):
            def _init_nested_resources(self):
                self.Child = CreateNestedResource(ChildResource, parent=self)
        parent = Parent('123456')
        self.assertEqual(parent.Child.__name__, 'ParentChild')


if __name__ == '__main__':
    unittest.main()



"""Exceptions"""


class ContiamoException(Exception):
    """Base class for the library's errors."""
    default_message = None
    include_error = False

    def __init__(self, message=None, http_body=None, http_status=None, json_body=None, headers=None):
        # if bytes, decode to string
        if http_body and hasattr(http_body, 'decode'):
            try:
                http_body = http_body.decode('utf-8')
            except BaseException:
                http_body = (
                    '<Could not decode body as utf-8. Please report to support@contiamo.com>')

        error_message = message if message else self.default_message
        if self.include_error and http_status:
            error_message += '\nAPI error %d: %s' % (http_status, http_body)
        super().__init__(error_message)

        self.http_body = http_body
        self.http_status = http_status
        self.json_body = json_body
        self.headers = headers or {}


class APIConnectionError(ContiamoException):
    """Basic connection errors such as server not responding."""
    pass


class AuthenticationError(ContiamoException):
    """Authentication error due to invalid credentials."""
    default_message = (
        'Authentication failed, please check your credentials (project id and API key) '
        'or contact us if the problem persists: support@contiamo.com')


class NotFoundError(ContiamoException):
    """Error raised when resource is not found after successful authentication."""
    default_message = 'The resource you requested was not found.'


class DataSourceError(ContiamoException):
    """Raises errors related to querying or updating data sources."""
    default_message = 'There was an error with the data source involved in your request.'
    include_error = True


class UpdateError(ContiamoException):
    """Errors related to saving or updating resources."""
    default_message = 'An error happened while creating or updating the resource.'
    include_error = True


class QueryError(ContiamoException):
    """Errors related to published query execution."""
    default_message = 'An error happened while executing the query.'
    include_error = True


class APIError(ContiamoException):
    """Catch-all error for all other non-200 responses including 500."""
    default_message = \
        'The server responded with an unusual error. Please retry or report the bug to support@contiamo.com'
    include_error = True


class ResponseError(ContiamoException):
    """Raised for invalid responses (e.g. invalid JSON); should be rare."""
    default_message = 'The response from the server was invalid. Please report the bug to support@contiamo.com'


class InvalidRequestError(ContiamoException):
    """Invalid request errors caught before sending the request."""
    pass



import vcr
import yaml

from . import utils
from contiamo.resources import Client


try:
    config = yaml.safe_load(open(utils.file_test_data('test_config.yml')))
except OSError:
    raise RuntimeError("Test config file not found. Email brandon@contiamo.com.")

api_base = config['api_base']
api_key = config['api_key']
project_id = config['project_id']
dashboard_id = config['dashboard_id']
app_id = config['app_id']
sql_table = config['sql_table']


class TestResources:

    client = Client(api_key, api_base=api_base)
    project = client.Project(project_id)
    dashboard = project.Dashboard.retrieve(dashboard_id)
    app = project.App(app_id)

    @vcr.use_cassette(utils.file_test_cassette('test_retrieve.yaml'))
    def test_retrieve(self):
        dashboard = self.project.Dashboard.retrieve(dashboard_id)
        assert isinstance(dashboard, self.project.Dashboard)
        assert dashboard['id'] == dashboard_id

    @vcr.use_cassette(utils.file_test_cassette('test_retrieve.yaml'))
    def test_fetch(self):
        dashboard = self.project.Dashboard(dashboard_id)
        dashboard.fetch()
        assert isinstance(dashboard, self.project.Dashboard)
        assert dashboard['id'] == dashboard_id

    @vcr.use_cassette(utils.file_test_cassette('test_get_list.yaml'))
    def test_get_list(self):
        dashboards = self.project.Dashboard.list()
        assert isinstance(dashboards[0], dict)
        dashboards = self.project.Dashboard.list(instantiate=True)
        assert isinstance(dashboards[0], self.project.Dashboard)

    @vcr.use_cassette(utils.file_test_cassette('test_create_dashboard.yaml'))
    def test_create_dashboard(self):
        dashboard_name = 'New dashboard'
        new_dashboard = self.project.Dashboard.create({'name': dashboard_name})
        assert isinstance(new_dashboard, self.project.Dashboard)
        assert new_dashboard['name'] == dashboard_name
        # Delete it so that we can create it next time
        new_dashboard.delete()

    @vcr.use_cassette(utils.file_test_cassette('test_create_contract.yaml'))
    def test_create_contract(self):
        contract_name = 'New contract'
        new_contract = self.app.Contract.create({'name': contract_name})
        assert isinstance(new_contract, self.app.Contract)
        assert new_contract['name'] == contract_name
        # Delete it so that we can create it next time
        new_contract.delete()

    @vcr.use_cassette(utils.file_test_cassette('test_modify.yaml'))
    def test_modify(self):
        new_layout = 'report'
        response = self.dashboard.modify({'layout_type': new_layout})
        assert response['layout_type'] == new_layout

    @vcr.use_cassette(utils.file_test_cassette('test_delete.yaml'))
    def test_delete(self):
        dashboard = self.project.Dashboard(dashboard_id)
        widget = dashboard.Widget.create(
            {"type": "text", "left": 8, "top": 1, "width": 15, "height": 2, "title": "", "description": ""})
        response = widget.delete()
        # assume that if server returns {} without error then delete was successful
        assert len(response) == 0

    @vcr.use_cassette(utils.file_test_cassette('test_sql_query.yaml'))
    def test_sql_query(self):
        result = self.project.query_sql(app_id, 'select * from %s limit 2;' % sql_table)
        assert result.loc[0, 'field_a'] == 1

    @vcr.use_cassette(utils.file_test_cassette('test_dynamic_query.yaml'))
    def test_dynamic_query(self):
        metric = "666571902:contract_data_value"
        dimension = "contract_data_category-category"
        payload = {
            "start_date": "2016-01", "end_date": "2016-01", "interval": "none", "period_unit": "month",
            "metrics": [{"key": metric}], "dimensions": [{"key": dimension}]
        }
        result = self.project.query(payload, use_column_names=False)
        assert metric in result and dimension in result

    # @TODO: add unit test for /data endpoints when API user is authorized



import unittest

from contiamo.dateparser import DateParser


class DateParseTestCase(unittest.TestCase):

    formats = [
        ['2014-12-31T23:59:59', 'second',  '2014-12-31T23:59:59'],
        ['2014-12-31T23:59',    'minute',  '2014-12-31T23:59:00'],
        ['2014-12-31T23',       'hour',    '2014-12-31T23:00:00'],
        ['2014-12-31',          'day',     '2014-12-31T00:00:00'],
        ['2014W01',             'week',    '2013-12-30T00:00:00'],
        ['2014W12',             'week',    '2014-03-17T00:00:00'],
        ['2014-12',             'month',   '2014-12-01T00:00:00'],
        ['2014Q1',              'quarter', '2014-01-01T00:00:00'],
        ['2014Q3',              'quarter', '2014-07-01T00:00:00'],
        ['2014',                'year',    '2014-01-01T00:00:00']
    ]

    def test_identification(self):
        for format in self.formats:
            parser = DateParser()
            self.assertEqual(parser.identifyPeriodUnit(format[0]), format[1])

    def test_parsing(self):
        for format in self.formats:
            parser = DateParser()
            parser.identifyPeriodUnit(format[0])
            date = parser.parse(format[0])
            self.assertEqual(date.strftime('%Y-%m-%dT%H:%M:%S'),
                             format[2], '{0} parsing failed'.format(format[1]))

    def test_formatting(self):
        for format in self.formats:
            parser = DateParser()
            parser.identifyPeriodUnit(format[0])
            date = parser.parse(format[0])
            self.assertEqual(parser.format(
                date), format[0], '{0} formatting failed'.format(format[1]))


if __name__ == '__main__':
    unittest.main()



from distutils.command.build_py import build_py
import os
from setuptools import setup, find_packages
import sys
import warnings

from pip._internal.download import PipSession
from pip._internal.req import parse_requirements


path, script = os.path.split(sys.argv[0])
os.chdir(os.path.abspath(path))

requirements = parse_requirements('requirements.txt', session=PipSession())
test_requirements = parse_requirements('requirements-dev.txt', session=PipSession())

requirements = [str(x.req) for x in requirements]
test_requirements = [str(x.req) for x in test_requirements]

if sys.version_info < (3,):
    warnings.warn(
        'Python 2 is not officially supported by Contiamo. '
        'If you have any questions, please file an issue on Github or '
        'contact us at support@contiamo.com.',
        DeprecationWarning)

# Don't import contiamo module here, since deps may not be installed
sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'contiamo'))
from version import VERSION

setup(
    name='contiamo',
    cmdclass={'build_py': build_py},
    version=VERSION,
    description='Contiamo Python client',
    author='Contiamo',
    author_email='support@contiamo.com',
    url='https://github.com/contiamo/contiamo-client-python',
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    test_suite='tests',
    tests_require=test_requirements,
    use_2to3=True,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ])



import unittest

from contiamo.utils import query_url_from_identifier, contract_url_template_from_identifier
from contiamo.errors import InvalidRequestError


class IdentifierToUrl(unittest.TestCase):
    api_base = 'https://api.some-name.com'

    # Query URLs
    def test_query_url(self):
        url, token = query_url_from_identifier(
            'query:olap:48590200:21237:randomquerytoken', api_base=self.api_base)
        self.assertEqual(url, self.api_base +
                         '/48590200/published_queries/21237/data.json')
        self.assertEqual(token, 'randomquerytoken')

    def test_sql_query_url(self):
        url, token = query_url_from_identifier(
            'query:sql:48590282:146:randomquerytoken', api_base=self.api_base)
        self.assertEqual(url, self.api_base +
                         '/48590282/published_queries/146/data.json')

    def test_backward_compatible_query_url(self):
        url, token = query_url_from_identifier(
            'query:48590200:21237:randomquerytoken', api_base=self.api_base)
        self.assertEqual(url, self.api_base +
                         '/48590200/published_queries/21237/data.json')

    # Contract URLs
    def test_contract_url_template(self):
        url_template = contract_url_template_from_identifier(
            'contract:48590121:666570779:test', api_base=self.api_base)
        self.assertEqual(
            url_template, self.api_base + '/48590121/apps/666570779/data_contracts/contracts/test/{action}')

    def test_invalid_contract_id(self):
        with self.assertRaises(InvalidRequestError):
            contract_url_template_from_identifier(
                'contract:48590121:666570779:test:randomcontracttoken', api_base=self.api_base)


if __name__ == '__main__':
    unittest.main()



from contiamo.http_client import HTTPClient
from contiamo.utils import query_url_from_identifier, parse_query_result, raise_response_error

import logging
logger = logging.getLogger(__name__)


def query(query_id, parse_dates=True, use_column_names=True, api_base='https://api.contiamo.com'):
    http_client = HTTPClient()
    url, token = query_url_from_identifier(query_id, api_base)
    response = http_client.request(
        'get', url, params={'resource_token': token})

    try:
        json_response = response.json()
    except ValueError as e:  # JSONDecodeError inherits from ValueError
        raise_response_error(e, response, logger)

    return parse_query_result(json_response, parse_dates=parse_dates, use_column_names=use_column_names)



import re
import pandas as pd
from isoweek import Week


class DateParser:
    """Inspects a date string to find out its date format and provides the right date parser"""

    def parse_second(date_string):
        return pd.datetime.strptime(date_string, '%Y-%m-%dT%H:%M:%S')

    def format_second(date):
        return date.strftime('%Y-%m-%dT%H:%M:%S')

    def parse_minute(date_string):
        return pd.datetime.strptime(date_string, '%Y-%m-%dT%H:%M')

    def format_minute(date):
        return date.strftime('%Y-%m-%dT%H:%M')

    def parse_hour(date_string):
        return pd.datetime.strptime(date_string, '%Y-%m-%dT%H')

    def format_hour(date):
        return date.strftime('%Y-%m-%dT%H')

    def parse_day(date_string):
        return pd.datetime.strptime(date_string, '%Y-%m-%d')

    def format_day(date):
        return date.strftime('%Y-%m-%d')

    def parse_week(date_string):
        split = date_string.split('W')
        date = Week(int(split[0]), int(split[1])).monday()
        return pd.datetime(date.year, date.month, date.day)

    def format_week(date):
        return Week.withdate(date).isoformat()

    def parse_month(date_string):
        return pd.datetime.strptime(date_string, '%Y-%m')

    def format_month(date):
        return date.strftime('%Y-%m')

    def parse_quarter(date_string):
        split = date_string.split('Q')
        return pd.datetime(int(split[0]), 1 + ((int(split[1])) - 1) * 3, 1)

    def format_quarter(date):
        return date.strftime('%Y') + 'Q' + str(((date.month - 1) // 3) + 1)

    def parse_year(date_string):
        return pd.datetime.strptime(date_string, '%Y')

    def format_year(date):
        return date.strftime('%Y')

    PERIOD_UNITS = [
        {
            'key': 'second',
            'regex': r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}$',
            'parser': parse_second,
            'formatter': format_second

        },
        {
            'key': 'minute',
            'regex': r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}$',
            'parser': parse_minute,
            'formatter': format_minute
        },
        {
            'key': 'hour',
            'regex': r'^\d{4}-\d{2}-\d{2}T\d{2}$',
            'parser': parse_hour,
            'formatter': format_hour
        },
        {
            'key': 'day',
            'regex': r'^\d{4}-\d{2}-\d{2}$',
            'parser': parse_day,
            'formatter': format_day
        },
        {
            'key': 'week',
            'regex': r'^\d{4}W\d{1,2}$',
            'parser': parse_week,
            'formatter': format_week
        },
        {
            'key': 'month',
            'regex': r'^\d{4}-\d{2}$',
            'parser': parse_month,
            'formatter': format_month
        },
        {
            'key': 'quarter',
            'regex': r'^\d{4}Q\d{1,2}$',
            'parser': parse_quarter,
            'formatter': format_quarter
        },
        {
            'key': 'year',
            'regex': r'^\d{4}$',
            'parser': parse_year,
            'formatter': format_year
        }
    ]

    def __init__(self):
        self.period_unit = None
        self.parser = None

    def identifyPeriodUnit(self, sample):
        for period_unit in self.PERIOD_UNITS:
            if re.search(period_unit['regex'], sample):
                self.period_unit = period_unit['key']
                self.parser = period_unit['parser']
                self.formatter = period_unit['formatter']
                break
        return self.period_unit

    def parse(self, date_string):
        return self.parser(date_string)

    def format(self, date):
        return self.formatter(date)
